#ifndef bqtQBlibErrorHH
#define bqtQBlibErrorHH

#undef B_ERROR
#define B_ERROR_INCLUDED

#include <qblib/types.hh>

#include <cstdio>
#include <vector>

struct B_ResumeRepeatEv { };
struct B_ResumeNextEv   { };
struct B_ErrorEv        { };
struct B_ResumeMainEv   { };

typedef struct { void* resumespot;
                 const char* file;
                 int line, col;
               } B_ERROR_PACKET;

extern void* ErrorHandler;
extern void* ActualResumeSpot;
extern void* StartResumingSpot;
extern int B_ERR_VALUE, B_ERL_VALUE;

static void* const ResumeRepeatKind = ((void*)(0UL));
static void* const ResumeNextKind   = ((void*)(1UL));

#define B_ERR_STMT_BEGIN() \
    { __label__ Repeat_Label; Repeat_Label: \
    try { try {

#define B_ERR_STMT_END(line, col, filename) \
              } \
              catch(const B_ErrorEv&) { __label__ Resume_Spot; \
                  static const B_ERROR_PACKET err = { &&Resume_Spot, filename, line,col }; \
                  goto* B_ERROR_Init(err); \
                Resume_Spot: \
                  B_RESUME_Dispatch(); \
    }         } \
    catch(const B_ResumeRepeatEv&) { goto Repeat_Label; } \
    catch(const B_ResumeNextEv&)   { } }

#define B_ERR_FRAME_MAIN_BEGIN() \
    Repeat_Main: try { \
        if(unlikely(ActualResumeSpot)) goto* ActualResumeSpot;

#define B_ERR_FRAME_MAIN_END() \
    } catch(const B_ResumeMainEv&) { goto Repeat_Main; }

#define B_ERR_FRAME_SUB_BEGIN() /*nothing to do */
#define B_ERR_FRAME_SUB_END()   /*nothing to do */



extern "C" {
void B_RESUME_FAILTEST();
void B_RESUME_Init();
void B_RESUMENEXT_Init();
void B_RESUMETO_Init(void*);
void B_RESUME_Dispatch() __attribute((noreturn));

#define B_RESUME()         do { B_RESUME_Init();          goto* StartResumingSpot; } while(0)
#define B_RESUME_NEXT()    do { B_RESUMENEXT_Init();      goto* StartResumingSpot; } while(0)
#define B_RESUME_TO(label) do { B_RESUMETO_Init(&&label); goto* StartResumingSpot; } while(0)

#define B_StmtERROR(code) do { /*throw "ERROR";*/ B_ERR_VALUE=(code); throw B_ErrorEv(); } while(0)
#define B_ERROR(code) B_StmtERROR(code)

#define ON_ERROR_GOTO(label) ErrorHandler=&&label
#define ON_ERROR_GOTO_NOWHERE() ErrorHandler=0

void* B_ERROR_Init(const B_ERROR_PACKET& err);
int B_ERROR_TranslateErrno();

#define B_ERROR_Errno() B_ERROR(B_ERROR_TranslateErrno())

static inline int B_FuncERR() { return B_ERR_VALUE; }
static inline int B_FuncERL() { return B_ERL_VALUE; }

}

#endif
